;; Generated package description from ack.el  -*- no-byte-compile: t -*-
(define-package "ack" "1.11" "interface to ack-like tools" 'nil :commit "90c90a80322aa9f26ce88f2f24a224afc4c994b8" :authors '(("Leo Liu" . "sdl.web@gmail.com")) :maintainer '("João Távora" . "joaotavora@gmail.com") :keywords '("tools" "processes" "convenience") :url "https://github.com/leoliu/ack-el")
